package it.mokabyte.pattern.dao;

import java.util.*;
import java.io.*;
import javax.naming.*;
import javax.ejb.*;

import it.mokabyte.pattern.pk.AccountPK;
import it.mokabyte.pattern.valueobject.AccountOM;
import it.mokabyte.pattern.exception.DAOException;
import it.mokabyte.pattern.dao.DaoConfig;
import it.mokabyte.pattern.util.SimpleTrace;

// import for XML parser
import org.w3c.dom.Document;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.DocumentBuilder;
import org.apache.xml.serialize.OutputFormat;
import org.apache.xml.serialize.Serializer;
import org.apache.xml.serialize.SerializerFactory;
import org.apache.xml.serialize.XMLSerializer;
import org.w3c.dom.NodeList;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;

//exception for parsering
import javax.xml.parsers.ParserConfigurationException;
import org.xml.sax.SAXException;

/**
 *  <p> Title:MyDaoXmlImpl </p>
 *
 *  <p>MyDaoXmlImpl makes an access to XML repository file
 *  whit this format:</p>
 *
 *  <p>
 *  <moka_account>
 *    <account pk="001">
 *      <name>Lucky</name>
 *      <address>Fortune 13</address>
 *      <balance>13456.67</balance>
 *    </account>
 *  </moka_account>
 *  </p>
 *
 * This file is in /temp/xml_src directory an  the name is
 * <b>reposiroty.xml<\b>
 *
 *
 *@author     S.Rossini, L.Dozio
 *@created    10 aprile 2002
 *@version    1.0
 */

public class MyDaoXmlImpl implements MyDao
{

 /**
  * the DOM document
  */
 private Document doc = null;
 /**
  * the repository file
  */
 private File fileRepository = null;
 /**
  * tag name
  */
 private String TAG_NAME = "name";
 /**
  * tag address
  */
 private String TAG_ADDRESS = "address";
 /**
  * tag balance
  */
 private String TAG_BALANCE = "balance";
 /**
  * tag account
  */
 private String TAG_ACCOUNT = "account";
 /**
  * tag primary key
  */
 private String ATTRIBUTE_PK = "pk";

 /**
  *  Constructor for the MyDaoXmlImpl object
  *
  *@exception  DAOException  there is not name file repository
  */
 public MyDaoXmlImpl() throws DAOException
 {
  SimpleTrace.trace(":MyDaoXmlImpl ... ");

  try{
      this.initResourceFile(DaoConfig.XML_FILE_NAME);
  } catch (Exception e){
      e.printStackTrace();
      throw new DAOException(e);
  }
 }

 /**
  * the Method inserts a new AccountOM in the repository
  *
  *@param  pk                the account primary key
  *@param  name              the account name
  *@param  address           the account email address
  *@param  balance           the account balance
  *@return                   the number of value inserted in the repository
  *@exception  DAOException  Object already exist, or problem with insert
  */
 public int insertAccount(AccountPK pk, String name, String address, double balance) throws DAOException
 {
  SimpleTrace.trace("MyDaoXmlImpl.insertAccount : START ");
  //insert a new element in the Xml file repository
  AccountOM account = new AccountOM(pk.id, name, address, balance);
  SimpleTrace.trace("MyDaoXmlImpl.insert: START - [" + account + "]");
  synchronized (this)
  {
   this.openDocument();
   if (this.getNode(pk.id) != null){
    SimpleTrace.trace("MyDaoXmlImpl.insert: Object ALREADY exist");
    this.doc = null;
    throw new DAOException("MyDaoXmlImpl.insert : Object ALREADY exist");
   }
   //Object not exist
   SimpleTrace.trace("MyDaoXmlImpl.insert: Objetc NOT Exist!");
   this.addNode(account);
   this.saveDocument();
  }
  SimpleTrace.trace("MyDaoXmlImpl.insert: END");
  return 1;
 }

 /**
  *  the Method deletes an exist account
  *
  *@param  pk                the  account primary key
  *@return                   the number of element deleted
  *@exception  DAOException  the account element not exist or problem with delete
  */
 public int deleteAccount(AccountPK pk) throws DAOException
 {
  SimpleTrace.trace("MyDaoXmlImpl.deleteAccount : START [" + pk + "]");
  synchronized (this)
  {
   this.openDocument();
   Node n = this.getNode(pk.id);
   if (n != null)
   {
    SimpleTrace.trace("MyDaoXmlImpl.deleteAccount: Objetc exist!");
    doc.getDocumentElement().removeChild(n);
   } else{
    this.doc = null;
    throw new DAOException("MyDaoXmlImpl.deleteAccount : Object Not Exist!!!");
   }
   this.saveDocument();
  }
  SimpleTrace.trace("MyDaoXmlImpl.deleteAccount : END");
  return 1;
 }

 /**
  *  the Method read an exist account
  *
  *@param  pk                the account primary key
  *@return                   Account value or an exception if value not exist
  *@exception  DAOException  the account not exist, or problem whit select
  */
 public AccountOM readAccount(AccountPK pk) throws DAOException
 {
  SimpleTrace.trace("MyDaoXmlImpl.readAccount : [" + pk + "]");
  AccountOM omResult = null;
  synchronized (this)
  {
   //open doc
   this.openDocument();
   Node n = this.getNode(pk.id);
   if (n != null)
   {
    omResult = this.getValueObject(n, pk.id);
   } else{
    this.doc = null;
    throw new DAOException("MyDaoXmlImpl.readAccount : Object NOT exist!!!");
   }
   this.saveDocument();
  }
  SimpleTrace.trace("MyDaoXmlImpl.readAccount : END [" + omResult + "]");
  return omResult;
 }

 /**
  *  the Method update an exist account
  *
  *@param  key               the account primary  key
  *@param  name              the account name value
  *@param  address           the account address value
  *@param  balance           the account balance value
  *@return                   the number of element updated
  *@exception  DAOException  account not exist or problem with update
  */
 public int updateAccount(AccountPK key, String name, String address, double balance) throws DAOException {
  SimpleTrace.trace("MyDaoXmlImpl.updateAccount : START ");
  //update the pk element in the Xml file repository);
  synchronized (this)
  {
   this.openDocument();
   Node n = this.getNode(key.id);
   if (n != null)
   {
    SimpleTrace.trace("MyDaoXmlImpl.updateAccount : Object exist!");
    AccountOM ovToUpdate = this.getValueObject(n, key.id);
    SimpleTrace.trace("update: name [" + name + "]");
    if ((name != null) || (!name.equals(""))){
     ovToUpdate.setName(name);
    }

    SimpleTrace.trace("MyDaoXmlImpl.updateAccount : address [" + address + "]");
    if ((address != null) || (!address.equals(""))){
     ovToUpdate.setAddress(address);
    }
    SimpleTrace.trace("MyDaoXmlImpl.updateAccount : balance [" + balance + "]");
    ovToUpdate.setBalance(balance);
    this.updateDocument(ovToUpdate);
    this.saveDocument();
   } else{
    SimpleTrace.trace("MyDaoXmlImpl.updateAccount : Object NOT exist!");
    this.doc = null;
    throw new DAOException("Object not exist!!!");
   }
  }
  SimpleTrace.trace("MyDaoXmlImpl.updateAccount : END");
  return 1;
 }

 /**
  *@param  key               the Primary Key for this account
  *@return                   Whether the indicated account exists
  *@exception  DAOException  thrown if an object not exist
  */
 public boolean accountExists(AccountPK key) throws DAOException
 {
  try
  {
    return (readAccount(key) != null);
  }
  catch (DAOException daoe)
  {
    throw daoe;
  }
 }

  /**
  *  init the XML reposistory
  *
  *@param  dataSourceName  the fiel name repository
  *@exception  Exception  not found the file in the specify path
  */
 private void initResourceFile(String dataSourceName) throws Exception
 {
  SimpleTrace.trace("MyDaoXmlImpl.initResourceFile : START [" + dataSourceName + "]");
  fileRepository = new File(dataSourceName);

  if (!fileRepository.exists()){
   throw new Exception("initResourceFile : file [" + fileRepository.getAbsolutePath() + "] NOT FOUND !");
  }
  SimpleTrace.trace("MyDaoXmlImpl.initResourceFile : END [" + dataSourceName + "]");
 }

 /**
  *  Add an account to the DOM document
  *
  *@param  accOm             the account value
  *@exception  DAOException  object already exist or problem with insert
  */
 private void addNode(AccountOM accOm) throws DAOException
 {
  SimpleTrace.trace("MyDaoXmlImpl.addNode : START [" + accOm + "]");
  try
  {
   if (this.getNode(accOm.getId())==null){
     //Object NOT exist
     org.w3c.dom.Element newNode = doc.createElement(TAG_ACCOUNT);
     newNode.setAttribute(ATTRIBUTE_PK, accOm.getId());
     newNode.appendChild(doc.createElement(TAG_NAME)).appendChild(doc.createTextNode(accOm.getName()));
     newNode.appendChild(doc.createElement(TAG_ADDRESS)).appendChild(doc.createTextNode(accOm.getAddress()));
     newNode.appendChild(doc.createElement(TAG_BALANCE)).appendChild(doc.createTextNode("" + accOm.getBalance()));
     doc.getDocumentElement().appendChild(newNode);
    }
    else throw new DAOException("MyDaoXmlImpl.addNode : Object ["+accOm.getId()+"] already exist!");
  } catch (Exception domExc){
   throw new DAOException(domExc);
  }
  SimpleTrace.trace("MyDaoXmlImpl.addNode: END ");
 }

 /**
  *  Get the valueObject attribute of the DOM repository
  *
  *@param  node     the node value
  *@param  pkValue  the primary key value
  *@return          The valueObject value
  */
 private AccountOM getValueObject(Node node, String pkValue){

  SimpleTrace.trace("MyDaoXmlImpl.getValueObject : START [" + node.getNodeName() + "] - [" + pkValue + "]");
  AccountOM accRes = new AccountOM();
  accRes.setCustId(pkValue);
  NodeList child = node.getChildNodes();
  //SimpleTrace.trace("getValueObject: [" + child.getLength() + "]");
  for (int i = 0; i < child.getLength(); i++)
  {
   Node n = child.item(i);
   //SimpleTrace.trace("getValueObject -> Node: name [" + n.getNodeName() + "] and the child: " +
   // "[" + n.getFirstChild() + "]");
   if (n.getNodeName().equals(TAG_NAME)){
    accRes.setName("" + n.getFirstChild());
   }
   if (n.getNodeName().equals(TAG_ADDRESS)){
    accRes.setAddress("" + n.getFirstChild());
   }
   if (n.getNodeName().equals(TAG_BALANCE)){
    accRes.setBalance(Double.parseDouble("" + n.getFirstChild()));
   }
  }
  SimpleTrace.trace("MyDaoXmlImpl.getValueObject: END [" + accRes + "]");
  return accRes;
 }


 /**
  *  Gets the node attribute of the DOM repository object
  *
  *@param  pkValue           the primary key
  *@return                   The node value or null if NOT exist
  *@exception  DAOException  problem with DOM select
  */
 private Node getNode(String pkValue) throws DAOException {

  SimpleTrace.trace("MyDaoXmlImpl.getNode : START [" + pkValue + "] - Node name : [" + doc.getDocumentElement().getNodeName() + "]");
  NodeList nl = doc.getDocumentElement().getChildNodes();
  //SimpleTrace.trace("getNode : childs number: [" + nl.getLength() + "]");
  for (int i = 0; i < nl.getLength(); i++)
  {
   NamedNodeMap nm = nl.item(i).getAttributes();
   if (nm != null)
   {
    //SimpleTrace.trace("getNode: attribute name [" + nm.item(0).getNodeName() + "]");
    if (nm.item(0).getNodeName().equals(ATTRIBUTE_PK)) {
     String pkV = nm.item(0).getNodeValue();
     //SimpleTrace.trace("getNode: pk value [" + pkV + "]");
     if (pkV.equals(pkValue)) {
      SimpleTrace.trace("MyDaoXmlImpl.getNode: FOUND!");
      return nl.item(i);
     }
    }
   }
  }
  SimpleTrace.trace("MyDaoXmlImpl.getNode : Element Not FOUND!");
  return null;
 }

 /**
  * the method update the value in the repository DOM obejct
  *
  *@param  ao                the account object to update
  *@exception  DAOException  the object not exist or problem with update
  */
 private void updateDocument(AccountOM ao) throws DAOException {
  SimpleTrace.trace("MyDaoXmlImpl.updateDocument : START [" + ao + "]");
  Node n = this.getNode(ao.getId());
  if (n==null) throw new DAOException("MyDaoXmlImpl.updateDocument : Object ["+ao.getId()+"] not exist!");
  NodeList nl = n.getChildNodes();
  //SimpleTrace.trace("updateDocument: " + nl.getLength());
  for (int i = 0; i < nl.getLength(); i++)
  {
   Node nUp = nl.item(i);
   if (nUp.getNodeName().equals(TAG_NAME)){
    nUp.getFirstChild().setNodeValue(ao.getName());
   }
   if (nUp.getNodeName().equals(TAG_ADDRESS)){
    nUp.getFirstChild().setNodeValue(ao.getAddress());
   }
   if (nUp.getNodeName().equals(TAG_BALANCE)){
    nUp.getFirstChild().setNodeValue("" + ao.getBalance());
   }
  }
  SimpleTrace.trace("MyDaoXmlImpl.updateDocument : END ");
 }

 /**
  *  the method inti the DOM document
  *
  *@exception  DAOException  problem with init DOM
  */
 private void openDocument() throws DAOException
 {
  SimpleTrace.trace("MyDaoXmlImpl.openDocument : START");
  if (doc == null){
   try
   {
    DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
    factory.setValidating(false);
    DocumentBuilder builder = factory.newDocumentBuilder();
    FileInputStream inFile = new FileInputStream(fileRepository.getAbsolutePath());
    doc = builder.parse(inFile);
    inFile.close();
   } catch (ParserConfigurationException pce){
      pce.printStackTrace();
      throw new DAOException(pce);
   } catch (SAXException sxe){
      sxe.printStackTrace();
      throw new DAOException(sxe);
   } catch (IOException ioe){
      ioe.printStackTrace();
      throw new DAOException(ioe);
   }
  }
  SimpleTrace.trace("MyDaoXmlImpl.openDocument : END");
 }


 /**
  *  the method save the values in an XML file repository
  *
  *@exception  DAOException  problem with saving
  */
 private void saveDocument() throws DAOException
 {
  FileOutputStream fout = null;
  SimpleTrace.trace("MyDaoXmlImpl.saveDocument : START");
  try
  {
   //Serializzazione del file e inseriemnto dei valori nel file fout
   fout = new FileOutputStream(fileRepository);
   OutputFormat format = new OutputFormat(doc);
   //Serialize DOM
   XMLSerializer serial = new XMLSerializer(fout, format);
   //Inizializzazione dell'out verso il file
   serial.asDOMSerializer();
   //As a DOM Serializer
   serial.serialize(doc.getDocumentElement());
   //Document serializzato che passa al file
   fout.flush();
   SimpleTrace.trace("MyDaoXmlImpl.saveDocument : END");
  }
  //end try
  catch (IOException ioe){
    ioe.printStackTrace();
    throw new DAOException(ioe);
  } catch (Exception ex){
    ex.printStackTrace();
    throw new DAOException(ex);
  } finally{
   if (fout != null){
    try{
     fout.close();
     doc = null;
    } catch (IOException ioe){ioe.printStackTrace();}
   }
  }
 }

}
